@@ -1,12 +1,12 @@
 /**
- * 
+ *
  * Ulfius Framework
- * 
+ *
  * REST framework library
- * 
+ *
  * u_response.c: response related functions defintions
- * 
- * Copyright 2015-2020 Nicolas Mora <mail@babelouest.org>
+ *
+ * Copyright 2015-2022 Nicolas Mora <mail@babelouest.org>
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public License
@@ -20,20 +20,15 @@
  *
  * You should have received a copy of the GNU General Public
  * License along with this library.  If not, see <http://www.gnu.org/licenses/>.
- * 
+ *
  */
 #include <string.h>
+#include <u_private.h>
+#include <ulfius.h>
 
-#include "u_private.h"
-#include "ulfius.h"
-
-/**
- * Add a cookie in the cookie map as defined in the RFC 6265
- * Returned value must be free'd after use
- */
 static char * ulfius_generate_cookie_header(const struct _u_cookie * cookie) {
   char * attr_expires = NULL, * attr_max_age = NULL, * attr_domain = NULL, * attr_path = NULL;
-  char * attr_secure = NULL, * attr_http_only = NULL, * cookie_header_value = NULL, * same_site = NULL;
+  char * attr_secure = NULL, * attr_http_only = NULL, * cookie_header_value = NULL, same_site[64] = {0};
 
   if (cookie != NULL) {
     if (cookie->expires != NULL) {
@@ -122,13 +117,13 @@ static char * ulfius_generate_cookie_header(const struct _u_cookie * cookie) {
     }
 
     if (cookie->same_site  == U_COOKIE_SAME_SITE_STRICT) {
-      same_site = o_strdup("; SameSite=Strict");
+      o_strcpy(same_site, "; SameSite=Strict");
     } else if (cookie->same_site == U_COOKIE_SAME_SITE_LAX) {
-      same_site = o_strdup("; SameSite=Lax");
-    } else {
-      same_site = o_strdup("");
+      o_strcpy(same_site, "; SameSite=Lax");
+    } else if (cookie->same_site == U_COOKIE_SAME_SITE_NONE) {
+      o_strcpy(same_site, "; SameSite=None");
     }
-    
+
     if (attr_expires == NULL || attr_max_age == NULL || attr_domain == NULL || attr_path == NULL || attr_secure == NULL || attr_http_only == NULL) {
       y_log_message(Y_LOG_LEVEL_ERROR, "Ulfius - Error allocating memory for ulfius_generate_cookie_header");
     } else {
@@ -143,25 +138,18 @@ static char * ulfius_generate_cookie_header(const struct _u_cookie * cookie) {
     o_free(attr_path);
     o_free(attr_secure);
     o_free(attr_http_only);
-    o_free(same_site);
     attr_expires = NULL;
     attr_max_age = NULL;
     attr_domain = NULL;
     attr_path = NULL;
     attr_secure = NULL;
     attr_http_only = NULL;
-    same_site = NULL;
     return cookie_header_value;
   } else {
     return NULL;
   }
 }
 
-/**
- * ulfius_set_response_header
- * adds headers defined in the response_map_header to the response
- * return the number of added headers, -1 on error
- */
 int ulfius_set_response_header(struct MHD_Response * response, const struct _u_map * response_map_header) {
   const char ** header_keys = u_map_enum_keys(response_map_header);
   const char * header_value;
@@ -181,11 +169,6 @@ int ulfius_set_response_header(struct MHD_Response * response, const struct _u_m
   return i;
 }
 
-/**
- * ulfius_set_response_cookie
- * adds cookies defined in the response_map_cookie
- * return the number of added headers, -1 on error
- */
 int ulfius_set_response_cookie(struct MHD_Response * mhd_response, const struct _u_response * response) {
   int ret;
   unsigned int i;
@@ -211,29 +194,15 @@ int ulfius_set_response_cookie(struct MHD_Response * mhd_response, const struct
   }
 }
 
-/**
- * ulfius_add_cookie_to_response
- * add a cookie to the cookie map
- * return U_OK on success
- */
-int ulfius_add_cookie_to_response(struct _u_response * response, const char * key, const char * value, const char * expires, const unsigned int max_age, 
+int ulfius_add_cookie_to_response(struct _u_response * response, const char * key, const char * value, const char * expires, const unsigned int max_age,
                                   const char * domain, const char * path, const int secure, const int http_only) {
-  return ulfius_add_same_site_cookie_to_response(response, key, value, expires, max_age, domain, path, secure, http_only, U_COOKIE_SAME_SITE_NONE);
+  return ulfius_add_same_site_cookie_to_response(response, key, value, expires, max_age, domain, path, secure, http_only, U_COOKIE_SAME_SITE_EMPTY);
 }
 
-/**
- * ulfius_add_same_site_cookie_to_response
- * add a cookie to the cookie map with a SameSite attribute
- * the same_site parameter must have one of the following values:
- * - U_COOKIE_SAME_SITE_NONE   - No SameSite attribute
- * - U_COOKIE_SAME_SITE_STRICT - SameSite attribute set to 'Strict'
- * - U_COOKIE_SAME_SITE_LAX    - SameSite attribute set to 'Lax'
- * return U_OK on success
- */
-int ulfius_add_same_site_cookie_to_response(struct _u_response * response, const char * key, const char * value, const char * expires, const unsigned int max_age, 
+int ulfius_add_same_site_cookie_to_response(struct _u_response * response, const char * key, const char * value, const char * expires, const unsigned int max_age,
                                             const char * domain, const char * path, const int secure, const int http_only, const int same_site) {
   unsigned int i;
-  if (response != NULL && key != NULL && (same_site == U_COOKIE_SAME_SITE_NONE || same_site == U_COOKIE_SAME_SITE_STRICT || same_site == U_COOKIE_SAME_SITE_LAX)) {
+  if (response != NULL && key != NULL && (same_site == U_COOKIE_SAME_SITE_EMPTY || same_site == U_COOKIE_SAME_SITE_NONE || same_site == U_COOKIE_SAME_SITE_STRICT || same_site == U_COOKIE_SAME_SITE_LAX)) {
     // Look for cookies with the same key
     for (i=0; i<response->nb_cookies; i++) {
       if (0 == o_strcmp(response->map_cookie[i].key, key)) {
@@ -265,7 +234,7 @@ int ulfius_add_same_site_cookie_to_response(struct _u_response * response, const
         }
       }
     }
-    
+
     // Key not found, inserting a new cookie
     if (response->nb_cookies == 0) {
       response->map_cookie = o_malloc(sizeof(struct _u_cookie));
@@ -289,7 +258,7 @@ int ulfius_add_same_site_cookie_to_response(struct _u_response * response, const
     response->map_cookie[response->nb_cookies].secure = secure;
     response->map_cookie[response->nb_cookies].http_only = http_only;
     response->map_cookie[response->nb_cookies].same_site = same_site;
-    if ((key != NULL && response->map_cookie[response->nb_cookies].key == NULL) || (value != NULL && response->map_cookie[response->nb_cookies].value == NULL) || 
+    if ((key != NULL && response->map_cookie[response->nb_cookies].key == NULL) || (value != NULL && response->map_cookie[response->nb_cookies].value == NULL) ||
         (expires != NULL && response->map_cookie[response->nb_cookies].expires == NULL) || (domain != NULL && response->map_cookie[response->nb_cookies].domain == NULL) ||
         (path != NULL && response->map_cookie[response->nb_cookies].path == NULL)) {
       y_log_message(Y_LOG_LEVEL_ERROR, "Ulfius - Error allocating memory for ulfius_add_cookie_to_response");
@@ -308,11 +277,6 @@ int ulfius_add_same_site_cookie_to_response(struct _u_response * response, const
   }
 }
 
-/**
- * ulfius_clean_cookie
- * clean the cookie's elements
- * return U_OK on success
- */
 int ulfius_clean_cookie(struct _u_cookie * cookie) {
   if (cookie != NULL) {
     o_free(cookie->key);
@@ -331,10 +295,6 @@ int ulfius_clean_cookie(struct _u_cookie * cookie) {
   }
 }
 
-/**
- * Copy the cookie source elements into dest elements
- * return U_OK on success
- */
 int ulfius_copy_cookie(struct _u_cookie * dest, const struct _u_cookie * source) {
   if (source != NULL && dest != NULL) {
     dest->key = o_strdup(source->key);
@@ -360,13 +320,6 @@ int ulfius_copy_cookie(struct _u_cookie * dest, const struct _u_cookie * source)
   return U_ERROR_PARAMS;
 }
 
-/**
- * ulfius_clean_response
- * clean the specified response's inner elements
- * user must free the parent pointer if needed after clean
- * or use ulfius_clean_response_full
- * return U_OK on success
- */
 int ulfius_clean_response(struct _u_response * response) {
   unsigned int i;
   if (response != NULL) {
@@ -388,6 +341,8 @@ int ulfius_clean_response(struct _u_response * response) {
     if ((struct _websocket_handle *)response->websocket_handle) {
       o_free(((struct _websocket_handle *)response->websocket_handle)->websocket_protocol);
       o_free(((struct _websocket_handle *)response->websocket_handle)->websocket_extensions);
+      pointer_list_clean_free(((struct _websocket_handle *)response->websocket_handle)->websocket_extension_list, &ulfius_free_websocket_extension_pointer_list);
+      o_free(((struct _websocket_handle *)response->websocket_handle)->websocket_extension_list);
       o_free(response->websocket_handle);
     }
 #endif
@@ -397,11 +352,6 @@ int ulfius_clean_response(struct _u_response * response) {
   }
 }
 
-/**
- * ulfius_clean_response_full
- * clean the specified response and all its elements
- * return U_OK on success
- */
 int ulfius_clean_response_full(struct _u_response * response) {
   if (ulfius_clean_response(response) == U_OK) {
     o_free(response);
@@ -411,11 +361,6 @@ int ulfius_clean_response_full(struct _u_response * response) {
   }
 }
 
-/**
- * ulfius_init_response
- * Initialize a response structure by allocating inner elements
- * return U_OK on success
- */
 int ulfius_init_response(struct _u_response * response) {
   if (response != NULL) {
     response->status = 200;
@@ -440,6 +385,7 @@ int ulfius_init_response(struct _u_response * response) {
     response->stream_user_data = NULL;
     response->timeout = 0;
     response->shared_data = NULL;
+    response->free_shared_data = NULL;
 #ifndef U_DISABLE_WEBSOCKET
     response->websocket_handle = o_malloc(sizeof(struct _websocket_handle));
     if (response->websocket_handle == NULL) {
@@ -454,6 +400,13 @@ int ulfius_init_response(struct _u_response * response) {
     ((struct _websocket_handle *)response->websocket_handle)->websocket_incoming_user_data = NULL;
     ((struct _websocket_handle *)response->websocket_handle)->websocket_onclose_callback = NULL;
     ((struct _websocket_handle *)response->websocket_handle)->websocket_onclose_user_data = NULL;
+    ((struct _websocket_handle *)response->websocket_handle)->rsv_expected = 0;
+    if ((((struct _websocket_handle *)response->websocket_handle)->websocket_extension_list = o_malloc(sizeof(struct _pointer_list))) == NULL) {
+      o_free(response->websocket_handle);
+      y_log_message(Y_LOG_LEVEL_ERROR, "Ulfius - Error allocating memory for response->websocket_handle->websocket_extension_list");
+      return U_ERROR_MEMORY;
+    }
+    pointer_list_init(((struct _websocket_handle *)response->websocket_handle)->websocket_extension_list);
 #endif
     return U_OK;
   } else {
@@ -461,11 +414,6 @@ int ulfius_init_response(struct _u_response * response) {
   }
 }
 
-/**
- * ulfius_copy_response
- * Copy the source response elements into the des response
- * return U_OK on success
- */
 int ulfius_copy_response(struct _u_response * dest, const struct _u_response * source) {
   unsigned int i;
   if (dest != NULL && source != NULL) {
@@ -494,8 +442,8 @@ int ulfius_copy_response(struct _u_response * dest, const struct _u_response * s
     } else {
       dest->map_cookie = NULL;
     }
-    
-    if (source->binary_body != NULL && source->binary_body_length > 0) {
+
+    if (source->binary_body != NULL && source->binary_body_length) {
       dest->binary_body = o_malloc(source->binary_body_length);
       if (dest->binary_body == NULL) {
         y_log_message(Y_LOG_LEVEL_ERROR, "Ulfius - Error allocating memory for dest->binary_body");
@@ -504,7 +452,7 @@ int ulfius_copy_response(struct _u_response * dest, const struct _u_response * s
       dest->binary_body_length = source->binary_body_length;
       memcpy(dest->binary_body, source->binary_body, source->binary_body_length);
     }
-    
+
     if (source->stream_callback != NULL) {
       dest->stream_callback = source->stream_callback;
       dest->stream_callback_free = source->stream_callback_free;
@@ -512,7 +460,7 @@ int ulfius_copy_response(struct _u_response * dest, const struct _u_response * s
       dest->stream_block_size = source->stream_block_size;
       dest->stream_user_data = source->stream_user_data;
     }
-    
+
     dest->shared_data = source->shared_data;
     dest->timeout = source->timeout;
 #ifndef U_DISABLE_WEBSOCKET
@@ -533,10 +481,6 @@ int ulfius_copy_response(struct _u_response * dest, const struct _u_response * s
   }
 }
 
-/**
- * create a new response based on the source elements
- * return value must be free'd after use
- */
 struct _u_response * ulfius_duplicate_response(const struct _u_response * response) {
   struct _u_response * new_response = NULL;
   if (response != NULL) {
@@ -560,12 +504,6 @@ struct _u_response * ulfius_duplicate_response(const struct _u_response * respon
   return new_response;
 }
 
-/**
- * ulfius_set_string_body_response
- * Set a string string_body to a response
- * string_body must end with a '\0' character
- * return U_OK on success
- */
 int ulfius_set_string_body_response(struct _u_response * response, const unsigned int status, const char * string_body) {
   if (response != NULL && string_body != NULL) {
     // Free all the bodies available
@@ -584,13 +522,8 @@ int ulfius_set_string_body_response(struct _u_response * response, const unsigne
   }
 }
 
-/**
- * ulfius_set_binary_body_response
- * Add a binary binary_body to a response
- * return U_OK on success
- */
 int ulfius_set_binary_body_response(struct _u_response * response, const unsigned int status, const char * binary_body, const size_t length) {
-  if (response != NULL && binary_body != NULL && length > 0) {
+  if (response != NULL && binary_body != NULL && length) {
     // Free all the bodies available
     o_free(response->binary_body);
     response->binary_body = NULL;
@@ -610,18 +543,13 @@ int ulfius_set_binary_body_response(struct _u_response * response, const unsigne
   }
 }
 
-/**
- * ulfius_set_empty_body_response
- * Set an empty response with only a status
- * return U_OK on success
- */
 int ulfius_set_empty_body_response(struct _u_response * response, const unsigned int status) {
   if (response != NULL) {
     // Free all the bodies available
     o_free(response->binary_body);
     response->binary_body = NULL;
     response->binary_body_length = 0;
-    
+
     response->status = status;
     return U_OK;
   } else {
@@ -629,14 +557,7 @@ int ulfius_set_empty_body_response(struct _u_response * response, const unsigned
   }
 }
 
-/**
- * ulfius_set_stream_response
- * Set an stream response with a status
- * set stream_size to -1 if unknown
- * set stream_block_size to a proper value based on the system
- * return U_OK on success
- */
-int ulfius_set_stream_response(struct _u_response * response, 
+int ulfius_set_stream_response(struct _u_response * response,
                                 const unsigned int status,
                                 ssize_t (* stream_callback) (void * stream_user_data, uint64_t offset, char * out_buf, size_t max),
                                 void (* stream_callback_free) (void * stream_user_data),
@@ -648,7 +569,7 @@ int ulfius_set_stream_response(struct _u_response * response,
     o_free(response->binary_body);
     response->binary_body = NULL;
     response->binary_body_length = 0;
-    
+
     response->status = status;
     response->stream_callback = stream_callback;
     response->stream_callback_free = stream_callback_free;
@@ -661,11 +582,6 @@ int ulfius_set_stream_response(struct _u_response * response,
   }
 }
 
-/**
- * ulfius_set_response_properties
- * Set a list of properties to a response
- * return U_OK on success
- */
 int ulfius_set_response_properties(struct _u_response * response, ...) {
   u_option option;
   int ret = U_OK;
@@ -678,7 +594,7 @@ int ulfius_set_response_properties(struct _u_response * response, ...) {
 
   if (response != NULL) {
     va_start(vl, response);
-    for (option = va_arg(vl, uint); option != U_OPT_NONE && ret == U_OK; option = va_arg(vl, uint)) {
+    for (option = va_arg(vl, unsigned int); option != U_OPT_NONE && ret == U_OK; option = va_arg(vl, unsigned int)) {
       switch (option) {
         case U_OPT_STATUS:
           response->status = va_arg(vl, long);
@@ -695,7 +611,7 @@ int ulfius_set_response_properties(struct _u_response * response, ...) {
         case U_OPT_AUTH_REALM:
           str_value = va_arg(vl, const char *);
           o_free(response->auth_realm);
-          if (o_strlen(str_value)) {
+          if (!o_strnullempty(str_value)) {
             response->auth_realm = o_strdup(str_value);
           } else {
             response->auth_realm = NULL;
@@ -735,6 +651,19 @@ int ulfius_set_response_properties(struct _u_response * response, ...) {
   return ret;
 }
 
+int ulfius_set_response_shared_data(struct _u_response * response, void * shared_data, void (* free_shared_data) (void * shared_data)) {
+  if (response != NULL && shared_data != NULL) {
+    if (response->free_shared_data != NULL && response->shared_data != NULL) {
+      response->free_shared_data(response->shared_data);
+    }
+    response->shared_data = shared_data;
+    response->free_shared_data = free_shared_data;
+    return U_OK;
+  } else {
+    return U_ERROR_PARAMS;
+  }
+}
+
 #ifndef U_DISABLE_JANSSON
 /**
  * ulfius_set_json_body_response
@@ -776,16 +705,47 @@ json_t * ulfius_get_json_body_response(struct _u_response * response, json_error
 }
 #endif
 
-/**
- * ulfius_add_header_to_response
- * add a header to the response
- * return U_OK on success
- */
 int ulfius_add_header_to_response(struct _u_response * response, const char * key, const char * value) {
   if (response != NULL && key != NULL && value != NULL) {
     return u_map_put(response->map_header, key, value);
-    return U_OK;
   } else {
     return U_ERROR_PARAMS;
   }
 }
+
+char * ulfius_export_response_http(const struct _u_response * response) {
+  char * out = NULL, * header;
+  const char * value = NULL, ** keys = NULL;
+  unsigned int i;
+
+  if (response != NULL) {
+    out = msprintf("HTTP/1.1 %ld\r\n", response->status);
+      
+    keys = u_map_enum_keys(response->map_header);
+    for (i=0; keys != NULL && keys[i] != NULL; i++) {
+      value = u_map_get(response->map_header, keys[i]);
+      if (value != NULL) {
+        out = mstrcatf(out, "%s: %s\r\n", keys[i], value);
+      } else {
+        out = mstrcatf(out, "%s:\r\n", keys[i]);
+      }
+    }
+
+    if (response->nb_cookies && !u_map_has_key_case(response->map_header, "Set-Cookie")) {
+      for (i=0; i<response->nb_cookies; i++) {
+        header = ulfius_generate_cookie_header(&response->map_cookie[i]);
+        out = mstrcatf(out, "Set-Cookie: %s\r\n", header);
+        o_free(header);
+      }
+    }
+    if (response->binary_body_length) {
+      out = mstrcatf(out, "Content-Length: %zu\r\n", response->binary_body_length);
+    }
+    out = mstrcatf(out, "\r\n");
+      
+    if (response->binary_body_length) {
+      out = mstrcatf(out, "%.*s\r\n", response->binary_body_length, response->binary_body);
+    }
+  }
+  return out;
+}
